import pytest

from xpinyin import Pinyin
from xpinyin.combs import _get_comb_indexes, get_combs


@pytest.fixture(scope='module')
def pinyin():
    return Pinyin()


def test_get_pinyin_with_default_splitter(pinyin):
    assert pinyin.get_pinyin('上海') == 'shang-hai'


def test_get_pinyin_with_splitter(pinyin):
    assert pinyin.get_pinyin('上海', splitter='') == 'shanghai'


def test_get_pinyin_mixed_words(pinyin):
    assert pinyin.get_pinyin('Apple发布iOS7', splitter='-') == 'Apple-fa-bu-iOS7'


def test_get_pinyin_with_tone_marks(pinyin):
    assert pinyin.get_pinyin('上海', tone_marks='marks') == 'sh\xe0ng-h\u01cei'
    assert pinyin.get_pinyin('秋', tone_marks='marks') == 'qiū'


def test_get_initial(pinyin):
    assert pinyin.get_initial('你') == 'N'


def test_get_initials(pinyin):
    assert pinyin.get_initials('你好') == 'N-H'


def test_get_initials_with_splitter(pinyin):
    assert pinyin.get_initials('你好', ' ') == 'N H'
    assert pinyin.get_initials('你好', '') == 'NH'


def test_get_initials_with_retroflex(pinyin):
    assert pinyin.get_initials('上海', '-', True) == 'SH-H'
    assert pinyin.get_initials('知识', '-', True) == 'ZH-SH'
    assert pinyin.get_initials('文字', '-', True) == 'W-Z'


# --- testing combinations auxiliary functions ---

def test_get_comb_indexes(pinyin):
    assert [[0, 0, 0], [0, 1, 0], [1, 0, 0], [1, 1, 0]] == _get_comb_indexes([2, 2, 1])


def test_get_comb_indexes_max_num(pinyin):
    assert [[0, 0, 0], [0, 1, 0], [1, 0, 0]], _get_comb_indexes([2, 2, 1], 3)


def test_get_combs(pinyin):
    assert ['a1@', 'a1#', 'a2@', 'a2#', 'b1@', 'b1#', 'b2@', 'b2#'] == get_combs([['a', 'b'], ['1', '2'], ['@', '#']])


def test_get_combs_splitter_max_num(pinyin):
    assert ['a 1 @', 'a 1 #', 'a 2 @', 'a 2 #', 'b 1 @'] == get_combs([['a', 'b'], ['1', '2'], ['@', '#']],
                                                                      splitter=' ', n=5)


def test_get_combs_max_num_too_big(pinyin):
    assert ['a||1||@', 'a||1||#', 'a||2||@', 'a||2||#', 'b||1||@', 'b||1||#', 'b||2||@', 'b||2||#'] == get_combs(
        [['a', 'b'], ['1', '2'], ['@', '#']], splitter='||', n=100)


# --- testing pinyin combinations ---

def test_get_pinyins_with_default_splitter(pinyin):
    assert pinyin.get_pinyins('上海') == ['shang-hai']


def test_get_pinyins_single_char(pinyin):
    # 4E50	LE4 YUE4 YAO4 LUO4 LIAO2
    assert ['lè', 'yuè', 'yào', 'luò', 'liáo'] == pinyin.get_pinyins('乐', splitter='', tone_marks='marks')


def test_get_pinyins_two_chars(pinyin):
    combs1 = pinyin.get_pinyins('音', splitter='', tone_marks='marks')
    combs2 = pinyin.get_pinyins('乐', splitter='', tone_marks='marks')
    combs12 = pinyin.get_pinyins('音乐', splitter='', tone_marks='marks')
    assert len(combs12) == len(combs1) * len(combs2)
    assert 'yīnyuè' in combs12


def test_get_pinyins_no_tones_uniq(pinyin):
    assert ['ma'] == pinyin.get_pinyins('吗', splitter='', tone_marks=None)


def test_get_pinyins_max_num(pinyin):
    assert 5 == len(pinyin.get_pinyins('音乐', splitter='', n=5))


def test_get_pinyins_mixed_words(pinyin):
    assert pinyin.get_pinyins('ABC串123', splitter=' ', tone_marks='marks') == ['ABC chuàn 123', 'ABC guàn 123']


def test_get_pinyins_long_seq(pinyin):
    text = """汉语拼音（Hànyǔ Pīnyīn），
        簡稱拼音，是一種以拉丁字母作普通话（現代標準漢語）標音的方案，為中文羅馬拼音的國際標準規範。
        汉语拼音在中国大陆作为基础教育内容全面使用，是义务教育的重要内容。在海外，特别是常用現代標準漢語的地区如新加坡、
        马来西亚、菲律宾和美国唐人街等，目前也在汉语教育中进行汉语拼音教学。臺灣自2008年開始，
        中文譯音使用原則也採用漢語拼音[1]，但舊護照姓名和部分地名、道路名稱仍採用威妥瑪拼音、
        郵政式拼音、國語羅馬字、國音二式抑或通用拼音[2]。"""
    assert 20 == len(pinyin.get_pinyins(text, n=20))
    assert 10 == len(pinyin.get_pinyins(text))  # limited to 10 by default



import re
from pathlib import Path
from typing import List, Optional

from xpinyin.combs import get_combs

PinyinToneMark = {
    0: "aoeiuv\u00fc",
    1: "\u0101\u014d\u0113\u012b\u016b\u01d6\u01d6",
    2: "\u00e1\u00f3\u00e9\u00ed\u00fa\u01d8\u01d8",
    3: "\u01ce\u01d2\u011b\u01d0\u01d4\u01da\u01da",
    4: "\u00e0\u00f2\u00e8\u00ec\u00f9\u01dc\u01dc",
}


class Pinyin:
    """Translate Chinese hanzi to pinyin (拼音) by Python, 汉字转拼音

    Usage
    -----
    ::

        >>> from xpinyin import Pinyin
        >>> p = Pinyin()
        >>> # default splitter is `-`
        >>> p.get_pinyin("上海")
        'shang-hai'
        >>> # show tone marks
        >>> p.get_pinyin("上海", tone_marks='marks')
        'shàng-hǎi'
        >>> p.get_pinyin("上海", tone_marks='numbers')
        >>> 'shang4-hai3'
        >>> # remove splitter
        >>> p.get_pinyin("上海", '')
        'shanghai'
        >>> # set splitter as whitespace
        >>> p.get_pinyin("上海", ' ')
        'shang hai'
        >>> p.get_initial("上")
        'S'
        >>> p.get_initials("上海")
        'S-H'
        >>> p.get_initials("上海", '')
        'SH'
        >>> p.get_initials("上海", ' ')
        'S H'
        >>> # get_initials with retroflex, #39
        >>> p.get_initials("上海", splitter='-', with_retroflex=True)
        'SH-H'
        >>> # get combinations of the multiple readings of the characters
        >>> p.get_pinyins('模型', splitter=' ', tone_marks='marks')
        ['mó xíng', 'mú xíng']
        >>> p.get_pinyins('模样', splitter=' ', tone_marks='marks')
        ['mó yáng', 'mó yàng', 'mó xiàng', 'mú yáng', 'mú yàng', 'mú xiàng']
    """

    data_path = Path(__file__).resolve().with_name('Mandarin.dat')

    def __init__(self, data_path: str = str(data_path)) -> None:
        lines = Path(data_path).read_text().splitlines()
        self.pinyins = dict(tuple(line.split('\t', maxsplit=1)) for line in lines)

    @staticmethod
    def decode_pinyin(s: str) -> str:
        s = s.lower()
        r = ""
        t = ""
        for c in s:
            if "a" <= c <= 'z':
                t += c
            elif c == ':':
                assert t[-1] == 'u'
                t = t[:-1] + "\u00fc"
            else:
                if '0' <= c <= '5':
                    tone = int(c) % 5
                    if tone != 0:
                        m = re.search("[aoeiuv\u00fc]+", t)
                        if m is None:
                            # pass when no vowels find yet
                            t += c
                        elif len(m.group(0)) == 1:
                            # if just find one vowels, put the mark on it
                            t = t[:m.start(0)] \
                                + PinyinToneMark[tone][PinyinToneMark[0].index(m.group(0))] \
                                + t[m.end(0):]
                        else:
                            # mark on vowels which search with "a, o, e" one by one
                            # when "i" and "u" stand together, make the vowels behind
                            for num, vowels in enumerate(("a", "o", "e", "ui", "iu")):
                                if vowels in t:
                                    t = t.replace(vowels[-1], PinyinToneMark[tone][num])
                                    break
                r += t
                t = ""
        r += t
        return r

    @staticmethod
    def convert_pinyin(word: str, convert: str) -> str:
        if convert == 'capitalize':
            return word.capitalize()
        if convert == 'lower':
            return word.lower()
        if convert == 'upper':
            return word.upper()

    def get_pinyins(self, chars: str, splitter: str = '-',
                    tone_marks: Optional[str] = None, convert: str = 'lower', n: int = 10) -> List[str]:
        """
        Get All pinyin combinations given all possible readings of each character.
        The number of combinations is limited par default to 10 to avoid exponential explosion on long texts.
        """
        all_pinyin_options = []  # a list of lists that we'll fill with all pinyin options for each character
        is_in_list = True  # in the list (otherwise, probably not a Chinese character)
        for char in chars:
            key = f"{ord(char):X}"
            if key not in self.pinyins:
                if is_in_list:
                    all_pinyin_options.append([char])  # add as is
                    is_in_list = False  # within a sequence of non Chinese characters
                else:
                    all_pinyin_options[-1][-1] += char  # add to previous sequence of non Chinese chars
            else:
                if tone_marks is None:  # in this case we may have duplicates if the variations differ just by the tones
                    char_py_options = []
                    for v in self.pinyins[key].split():
                        if v[0:-1] not in char_py_options:  # we remove the tone mark while we're at it
                            char_py_options.append(v[0:-1])
                else:
                    char_py_options = self.pinyins[key].split()
                last = 1 if n == 1 else len(char_py_options)
                if tone_marks == 'marks':
                    char_options = [Pinyin.decode_pinyin(o) for o in char_py_options[0:last]]
                else:  # 'numbers' or None
                    char_options = [o for o in char_py_options[0:last]]

                all_pinyin_options.append([Pinyin.convert_pinyin(c, convert) for c in char_options])
                is_in_list = True

        return get_combs(options=all_pinyin_options, splitter=splitter, n=n)

    def get_pinyin(self, chars: str, splitter: str = '-',
                   tone_marks=None, convert: str = 'lower') -> str:

        return self.get_pinyins(chars, splitter=splitter, tone_marks=tone_marks, convert=convert, n=1)[0]

    def get_initial(self, char: str, with_retroflex: bool = False) -> str:
        try:
            first_pinyin = self.pinyins[f"{ord(char):X}"].split(" ")[0]
            if with_retroflex and (first_pinyin[:2] in ('ZH', 'CH', 'SH')):
                return first_pinyin[:2]
            else:
                return first_pinyin[0]
        except KeyError:
            return char

    def get_initials(self, chars: str, splitter: str = '-', with_retroflex: bool = False):
        return splitter.join([self.get_initial(char, with_retroflex=with_retroflex) for char in chars])



from typing import List


def _get_comb_indexes(num_options_list: List[int], n=None) -> List[List[int]]:
    """
    Given a list with the number of possible options per place, returns a list of numbers representing combinations.
    The combinations are created via additions to a multi-radix number, from left to right
    (i.e. from smaller to larger numbers).

    e.g. [2, 2, 1] -> [[0, 0, 0], [0, 1, 0], [1, 0, 0], [1, 1, 0]]
    i.e. we have 2 options (0, 1) for the first and second places and one option (0) for the third.

    :param num_options_list: a list with the number of options per place
    :param n: The maximal number of requested combinations. If None, all possible combinations will be returned
    """
    # calculate the maximal number of possible combinations
    n_max = 1
    for j in num_options_list:
        n_max *= j
    n = min(n, n_max) if n is not None else n_max
    if n == 0:
        raise ValueError("Can't create combinations with 0-length lists")

    n_items = len(num_options_list)
    curr = [0] * n_items
    combs = [list.copy(curr)]
    i = n_items - 1
    count = 1
    while count < n:
        curr[i] = (curr[i] + 1) % num_options_list[i]
        if curr[i] != 0:
            combs.append(list.copy(curr))
            count += 1
            i = n_items - 1  # reset to right-most digit
        else:
            i -= 1  # try previous (left) digit

    return combs


def get_combs(options: List[List[str]], splitter: str = '', n: int = None) -> List[str]:
    """
    Given a list of options per place, returns up to n combinations
    e.g.: [['a'], ['1' ,'2'], ['@']] -> [a1@, a2@]
    For instance, ['1' ,'2'] is the group defining the options for the second place

    :param options: a list with a list of options for each group.
    :param splitter: a string to separate the groups
    :param n: The maximal number of requested combinations. If None, all possible combinations will be returned
    """
    combs = []
    comb_numbers = [len(o) for o in options]
    combs_indexes = _get_comb_indexes(comb_numbers, n)

    for c in combs_indexes:  # e.g. [0,2,1]
        comb = []
        for i in range(len(c)):
            comb.append(options[i][c[i]])
        combs.append(splitter.join(comb))

    return combs



from pathlib import Path

from setuptools import find_packages, setup

here = Path(__file__).resolve().parent
README = (here / "README.rst").read_text()
CHANGES = (here / "CHANGES.rst").read_text()

setup(
    name="xpinyin",
    version="0.7.7",
    description="Translate Chinese hanzi to pinyin (拼音) by Python, 汉字转拼音",
    long_description=README + "\n\n" + CHANGES,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="pinyin",
    author="Eric Lo",
    author_email="lxneng@gmail.com",
    url="https://github.com/lxneng/xpinyin",
    packages=find_packages("src"),
    test_suite="xpinyin.tests",
    package_dir={"": "src"},
    include_package_data=True,
    extras_require={
        "test": ["pytest>=6.2.0"],
    },
    license="MIT",
)
